#![cfg(target_os = "wasi")]
use std::sync::{Arc, Mutex};
use regex::Regex;
use rustls_pki_types::ServerName;
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader, ReadHalf, WriteHalf};
use tokio::net::TcpStream;
use tokio_rustls::rustls::{ClientConfig, RootCertStore};
use std::convert::TryFrom;
use std::time::Duration;
use anyhow::anyhow;
use imap_proto::{self, MailboxDatum};
use imap_proto::parser::parse_response;
use tokio::time::{sleep, timeout};

pub struct Client{
    reader: Arc<tokio::sync::Mutex<BufReader<ReadHalf<tokio_rustls::client::TlsStream<TcpStream>>>>>,
    writer: Arc<tokio::sync::Mutex<WriteHalf<tokio_rustls::client::TlsStream<TcpStream>>>>
}

impl Client {
    pub fn new(stream: tokio_rustls::client::TlsStream<tokio::net::TcpStream>) -> Client {
        let (reader, writer) = tokio::io::split(stream);
        let reader = Arc::new(tokio::sync::Mutex::new(BufReader::new(reader)));
        let writer = Arc::new(tokio::sync::Mutex::new(writer));
        
        Client {
            reader,
            writer,
        }
    }

    async fn run_command(&mut self, command: String) -> anyhow::Result<()>{
        let writer_clone = Arc::clone(&self.writer);
        let mut writer = writer_clone.lock().await;
        
        (*writer).write_all((command + "\r\n").as_bytes()).await?;
        (*writer).flush().await?;
        drop(writer);

        Ok(())
    }

    pub async fn read_response_by_id(&mut self, id: String) -> anyhow::Result<String>{
        let mut response = String::new();
        let mut temp = String::new();
        let reader_clone = Arc::clone(&self.reader);
        let mut reader = reader_clone.lock().await;

        while let Ok(bytes) = (*reader).read_line(&mut temp).await {
            println!("Temp: {:?}", temp);
            if bytes == 0 || temp.contains(&*id.to_string()) { // if bytes == 0 || temp.contains(&*format!("{} OK", id))
                break;
            }
            response += &*temp;
            temp.clear();
        }
        drop(reader);

        Ok(response)
    }

    pub async fn read_response_by_keyword(&self, key: String) -> anyhow::Result<String> {
        // Clone the Arc to move into the async task
        let reader_clone = Arc::clone(&self.reader);

        let handler = tokio::spawn(async move {
            let mut response = String::new();
            let mut temp = String::new();
            loop {
                // Lock the reader for access
                let mut reader = reader_clone.lock().await;

                // Use tokio::select! for concurrent operations
                let result = timeout(Duration::from_secs(2), reader.read_line(&mut temp)).await;
                match result {
                    Ok(Ok(bytes)) => {
                        println!("Temp: {:?}", temp);
                        if bytes == 0 || temp.contains(&*key) {
                            return Ok(response)
                        }
                        response += &*temp;
                        temp.clear();
                    }
                    Ok(Err(e)) => {
                        eprintln!("Error reading line: {:?}", e);
                        return Err(anyhow!("Error reading line from IMAP server"));
                    }
                    _ => {
                        sleep(Duration::from_secs(1)).await;
                        tokio::task::yield_now().await;
                    }
                }
                
                drop(reader);
            }
        });

        // Wait for the handler to complete and return the result
        handler.await.unwrap_or_else(|e| Err(anyhow!("Task panicked: {:?}", e)))
    }

    pub async fn login(&mut self, username: String, password: String) -> anyhow::Result<String>{
        let id = "aLogin".to_string();
        self.run_command(format!("{} LOGIN {} {}", id, username, password)).await?;

        Ok(self.read_response_by_id(id).await?)
    }

    pub async fn logout(&mut self) -> anyhow::Result<()>{
        self.run_command("LOGOUT".to_string()).await.expect("Unable to run a command");

        Ok(())
    }

    pub async fn list(&mut self) -> anyhow::Result<Vec<String>>{
        let list_command = "aList LIST \"\" \"*\"";
        self.run_command(list_command.to_string()).await.expect("Unable to run a command");
        let lists = self.read_response_by_id("aList".to_string()).await?;
        let mut lines = lists.as_bytes();
        
        let mut things = Vec::new();
        loop {
            if lines.is_empty() {
                break;
            }

            match parse_response(lines) {
                Ok((rest, resp)) => {
                    lines = rest;
                    things.push(resp)
                }
                _ => {
                    break;
                }
            }
        }
        
        Ok(things.into_iter()
            .filter_map(|datum| {
                if let imap_proto::Response::MailboxData(MailboxDatum::List { name, .. }) = datum {
                    Some(name.to_string())
                } else {
                    None
                }
            })
            .collect::<Vec<String>>())
    }

    pub async fn select(&mut self, list: String) -> anyhow::Result<String>{
        let id = "aSelect".to_string();
        self.run_command(format!("{} SELECT {}", id, list)).await.expect("Unable to run a command");

        self.read_response_by_id(id).await
    }


    pub async fn uid_search(&mut self, query: String) -> anyhow::Result<Vec<String>> {
        let id = "aUidSearch".to_string(); // Tag for the command
        self.run_command(format!("{} UID SEARCH {}", id, query)).await.expect("Unable to run a command");

        Ok(Self::uids_parse(self.read_response_by_id(id).await?).unwrap())
    }

    fn uids_parse(uids: String) -> anyhow::Result<Vec<String>>{
        let re = Regex::new(r"\b\d+\b")?;
        let uids = re.find_iter(&*uids)
            .filter_map(|mat| mat.as_str().parse::<String>().ok())
            .collect::<Vec<String>>();

        Ok(uids)
    }

    pub async fn uid_fetch(&mut self, uid: String, query: String) -> anyhow::Result<String>{
        let id = "aUidFetch".to_string();
        self.run_command(format!("{} UID FETCH {} {}", id, uid, query)).await.expect("Unable to run a command");

        Ok(self.read_response_by_id(id).await?)
    }
    
    pub async fn mark_deleted(&mut self, uid: u32) -> anyhow::Result<()>{
        let id = "aMarkDeleted".to_string();
        self.run_command(format!("{} UID STORE {} +FLAGS (\\Deleted)", id, uid)).await.unwrap();
        Ok(())
    }
    
    pub async fn expunge(&mut self) -> anyhow::Result<()>{
        let id = "aExpunge".to_string();
        self.run_command(format!("{} EXPUNGE", id)).await.unwrap();
        Ok(())
    }
    
    pub async fn create_folder(&mut self, name: String) -> anyhow::Result<()>{
        if name.is_empty() { return Err(anyhow!("Folder name is empty")) }
        let id = "aCreateFolder".to_string();
        self.run_command(format!("{} CREATE \"{}\"", id, name)).await.unwrap();
        Ok(())
    }

    pub async fn delete_folder(&mut self, name: String) -> anyhow::Result<()>{
        if name.is_empty() { return Err(anyhow!("Folder name is empty")) }

        let id = "aDeleteFolder".to_string();
        self.run_command(format!("{} DELETE \"{}\"", id, name)).await.unwrap();
        Ok(())
    }

    pub async fn rename_folder(&mut self, name: String, new_name: String) -> anyhow::Result<()>{
        if name.is_empty() { return Err(anyhow!("Folder name is empty")) }

        let id = "aRenameFolder".to_string();
        self.run_command(format!("{} RENAME \"{}\" \"{}\"", id, name, new_name)).await.unwrap();
        Ok(())
    }
    
    pub async fn idle(&mut self) -> anyhow::Result<String>{
        let id = "aIdle".to_string();
        self.run_command(format!("{} IDLE", id)).await.unwrap();
        Ok(id)
    }

    pub async fn idle_done(&mut self) -> anyhow::Result<String>{
        let id = "aIdleDone".to_string();
        self.run_command(format!("{} DONE", id)).await.unwrap();
        Ok(id)
    }
}

pub async fn connect_to_imap_server(domain: String, port: u16) -> anyhow::Result<Client> {
    let mut root_cert_store = RootCertStore::empty();
    root_cert_store.extend(webpki_roots::TLS_SERVER_ROOTS.iter().cloned());
    let config = ClientConfig::builder()
        .with_root_certificates(root_cert_store)
        .with_no_client_auth();
    let connector = tokio_rustls::TlsConnector::from(Arc::new(config));
    let dnsname = ServerName::try_from(domain.clone()).unwrap();

    // Connect to the IMAP server over SSL
    let stream = tokio::net::TcpStream::connect(&(domain, port)).await?;
    let stream = connector.connect(dnsname, stream).await?;

    Ok(Client::new(stream))
}